# =========================================================================
# Credit Classification Analysis - R Script
# Extracted from credit_classification_comparison.Rmd
# Author: Zhaopeng Qu
# Date: Generated from RMarkdown file
# =========================================================================

# Load Required Libraries
# =========================================================================
library(pacman)
p_load(
  tidyverse,   # for data wrangling and visualization
  tidymodels,  # for data modeling
  vip,         # for variable importance
  here,        # for referencing files and folders
  ranger,      # for random forest
  kknn,        # for k-nearest neighbors
  rpart,       # for decision trees
  rpart.plot,  # for plotting decision trees
  patchwork,   # for combining plots
  ISLR,
  broom,
  ggplot2, ggthemes, ggforce, ggridges, cowplot, scales,
  latex2exp, viridis, extrafont, gridExtra, plotly, ggformula,
  kableExtra, DT, huxtable,
  data.table, dplyr, snakecase, janitor,
  lubridate, knitr,
  caret, rpart, rpart.plot, rattle,
  here, magrittr, parallel,
  GGally, skimr, stringr, forcats, readr, janitor,
  # Additional packages for machine learning
  randomForest, e1071, gbm, xgboost, nnet, FNN,
  corrplot, VIM, elasticnet
)

# Set seed for reproducibility
set.seed(1203)

# Define colors
red_pink   = "#e64173"
turquoise  = "#20B2AA"
orange     = "#FFA500"
red        = "#fb6107"
blue       = "#3b3b9a"
green      = "#8bb174"
grey_light = "grey70"
grey_mid   = "grey50"
grey_dark  = "grey20"
purple     = "#6A5ACD"
slate      = "#314f4f"

# Data Loading and Preprocessing
# =========================================================================

# Read the credit data
credit_raw <- read_csv("data/credit.csv")

# Preprocessing steps from the original file
credit_processed <- credit_raw %>%
  rename(default = Default) %>% 
  mutate(default = as_factor(default)) %>% 
  mutate(history = fct_collapse(history,
    good     = c("A30", "A31"),
    poor     = c("A32", "A33"),
    terrible = c("A34")
  )) %>%
  mutate(foreign = fct_recode(foreign,
    "foreign" = "A201",
    "german"  = "A202"
  )) %>%
  mutate(purpose = fct_collapse(purpose,
    newcar       = c("A40"),
    usedcar      = c("A41"),
    goods_repair = c("A42", "A43", "A44", "A45"),
    edu          = c("A46", "A48"),
    na           = c("A47"),
    biz          = c("A49", "A410")
  )) %>% 
  mutate(rent = factor(housing == "A151"))

# Select subset of variables for analysis
credit <- credit_processed %>% 
  select(default, duration, amount,
         installment, age, history,
         purpose, foreign, rent)

# Display data structure
cat("Dataset shape:", nrow(credit), "rows,", ncol(credit), "columns\n")
glimpse(credit)

# Data Quality Assessment
# =========================================================================

# Check for missing values
missing_values <- colSums(is.na(credit))
cat("Missing values per column:\n")
print(missing_values)

# Check data types
cat("\nData types:\n")
print(sapply(credit, class))

# Exploratory Data Analysis 
# =========================================================================

# Check target variable distribution
p1 <- credit %>% 
  count(default) %>% 
  mutate(prop = n/sum(n)) %>% 
  ggplot(aes(x = default, y = prop)) +
  geom_col(fill = "navy", alpha = 0.7) +
  geom_text(aes(label = paste0(round(prop*100, 1), "%")), 
            vjust = -0.3, size = 4) +
  labs(title = "Distribution of Default Variable",
       x = "Default Status", y = "Proportion") +
  theme_minimal() +
  theme(plot.title = element_text(hjust = 0.5))

print(p1)

# Summary Statistics
# Create a custom summary with mean, standard deviation, max and min
summary_stats <- credit %>%
  select(where(is.numeric)) %>%  # Select only numeric variables
  summarise_all(list(
    Mean = ~round(mean(., na.rm = TRUE), 3),
    SD = ~round(sd(., na.rm = TRUE), 3),
    Max = ~round(max(., na.rm = TRUE), 3),
    Min = ~round(min(., na.rm = TRUE), 3)
  )) %>%
  pivot_longer(everything(), names_to = "Variable_Stat", values_to = "Value") %>%
  separate(Variable_Stat, into = c("Variable", "Statistic"), sep = "_") %>%
  pivot_wider(names_from = Statistic, values_from = Value)

print(kable(summary_stats, caption = "Descriptive Statistics for Numeric Variables"))

# Categorical Variables Analysis
# Function to calculate proportions for categorical variables
calculate_cat_proportions <- function(var_name, data) {
  data %>%
    count(!!sym(var_name)) %>%
    mutate(
      Variable = var_name,
      Category = as.character(!!sym(var_name)),
      Count = n,
      Proportion = round(n / sum(n), 3)
    ) %>%
    select(Variable, Category, Count, Proportion)
}

# Get categorical variables (excluding numeric ones)
categorical_vars <- credit %>%
  select(where(function(x) !is.numeric(x))) %>%
  names()

# Calculate proportions for all categorical variables
cat_stats <- map_dfr(categorical_vars, ~calculate_cat_proportions(.x, credit))

# Split categorical variables for better display
cat_vars_part1 <- c("default", "history")
cat_vars_part2 <- c("purpose", "foreign", "rent")

cat_stats_part1 <- cat_stats %>% filter(Variable %in% cat_vars_part1)
cat_stats_part2 <- cat_stats %>% filter(Variable %in% cat_vars_part2)

print(kable(cat_stats_part1, caption = "Categorical Variables (Part 1)"))
print(kable(cat_stats_part2, caption = "Categorical Variables (Part 2)"))

# Correlation Analysis
# Calculate correlation matrix for numeric variables
numeric_data <- credit %>% 
  select(where(is.numeric))

correlation_matrix <- cor(numeric_data)

# Create correlation plot
corrplot(correlation_matrix, 
         method = "color", 
         type = "upper", 
         order = "hclust", 
         tl.cex = 0.8, 
         tl.col = "black",
         title = "Correlation Matrix of Numeric Features",
         mar = c(0,0,2,0),
         addCoef.col = "black",
         number.cex = 0.8)

# Feature-Target Associations
# Calculate point-biserial correlation (numeric vs binary)
numeric_vars <- c("duration", "amount", "installment", "age")

# Convert default to numeric for correlation
credit_numeric <- credit %>%
  mutate(default_numeric = as.numeric(default) - 1)  # Convert to 0/1

pb_correlations <- numeric_vars %>%
  map_dbl(~cor(credit_numeric[[.x]], credit_numeric$default_numeric)) %>%
  round(3)

# Create correlation table
numeric_target_corr <- data.frame(
  Variable = numeric_vars,
  Correlation = pb_correlations,
  Interpretation = case_when(
    abs(pb_correlations) > 0.3 ~ "Strong",
    abs(pb_correlations) > 0.1 ~ "Moderate", 
    TRUE ~ "Weak"
  )
)

print(kable(numeric_target_corr, caption = "Numeric Variables vs Default Risk"))

# Categorical Variables vs Default using Cramér's V
# Function to calculate Cramér's V (categorical association measure)
cramers_v <- function(x, y) {
  # Create contingency table
  tbl <- table(x, y)
  # Calculate chi-square statistic
  chi2 <- chisq.test(tbl)$statistic
  # Calculate Cramér's V
  n <- sum(tbl)
  min_dim <- min(nrow(tbl), ncol(tbl)) - 1
  sqrt(chi2 / (n * min_dim))
}

# Calculate associations for categorical variables
cat_vars <- c("history", "purpose", "foreign", "rent")

cat_associations <- cat_vars %>%
  map_dbl(~cramers_v(credit[[.x]], credit$default)) %>%
  round(3)

# Create association table
cat_target_assoc <- data.frame(
  Variable = cat_vars,
  Cramers_V = cat_associations,
  Interpretation = case_when(
    cat_associations > 0.3 ~ "Strong",
    cat_associations > 0.1 ~ "Moderate",
    TRUE ~ "Weak"
  )
)

print(kable(cat_target_assoc, caption = "Cramér's V: Categorical Variables vs Default Risk"))

# Data Splitting and Preprocessing
# =========================================================================

# Train-test split (80-20)
credit_split <- credit %>% initial_split(prop = 0.8, strata = default)
credit_train <- credit_split %>% training()
credit_test  <- credit_split %>% testing()

# Cross-validation folds
credit_folds <- credit_train %>% vfold_cv(v = 5, strata = default)

cat("Training set size:", nrow(credit_train), "\n")
cat("Test set size:", nrow(credit_test), "\n")
cat("Cross-validation folds: 5\n")
cat("Stratified sampling: Yes\n")

# Create preprocessing recipe for standardization
scaling_recipe <- recipe(default ~ ., data = credit_train) %>%
  step_dummy(all_nominal_predictors()) %>%
  step_zv(all_predictors()) %>%
  step_normalize(all_numeric_predictors())

# Prep and apply the recipe
scaling_prep <- prep(scaling_recipe)
credit_train_scaled <- bake(scaling_prep, new_data = credit_train)
credit_test_scaled <- bake(scaling_prep, new_data = credit_test)

# Check scaling results
numeric_cols <- credit_train_scaled %>% select(where(is.numeric)) %>% names()
train_means <- sapply(credit_train_scaled[numeric_cols], mean, na.rm = TRUE)
train_sds <- sapply(credit_train_scaled[numeric_cols], sd, na.rm = TRUE)

cat("Feature scaling completed using tidymodels recipes\n")
cat("Training features mean (should be ~0):", round(mean(train_means), 6), "\n")
cat("Training features std (should be ~1):", round(mean(train_sds), 3), "\n")
cat("Number of dummy variables created:", 
    ncol(credit_train_scaled) - ncol(credit_train), "\n")

# Model Specifications
# =========================================================================

# 1. Logistic Regression (no penalty)
logistic_spec <- 
  logistic_reg() %>%
  set_engine("glm") %>%
  set_mode("classification")

# 2. Lasso Regression (L1 penalty)
lasso_spec <- 
  logistic_reg(penalty = tune(), mixture = 1) %>%
  set_engine("glmnet") %>%
  set_mode("classification")

# 3. Ridge Regression (L2 penalty)
ridge_spec <- 
  logistic_reg(penalty = tune(), mixture = 0) %>%
  set_engine("glmnet") %>%
  set_mode("classification")

# 4. Decision Tree
tree_spec <- 
  decision_tree(
    cost_complexity = tune(),
    tree_depth = tune(),
    min_n = tune()
  ) %>%
  set_engine("rpart") %>%
  set_mode("classification")

# 5. Random Forest
rf_spec <- 
  rand_forest(
    mtry = tune(),
    trees = 1000,
    min_n = tune()
  ) %>%
  set_engine("ranger", importance = "impurity") %>%
  set_mode("classification")

# 6. K-Nearest Neighbors
knn_spec <- 
  nearest_neighbor(
    neighbors = tune(),
    weight_func = tune(),
    dist_power = tune()
  ) %>%
  set_engine("kknn") %>%
  set_mode("classification")

cat("✓ All 6 model specifications created:\n")
cat("  1. Logistic Regression (GLM)\n")
cat("  2. Lasso Regression (GLMNet, L1)\n")
cat("  3. Ridge Regression (GLMNet, L2)\n")
cat("  4. Decision Tree (rpart)\n")
cat("  5. Random Forest (ranger)\n")
cat("  6. K-Nearest Neighbors (kknn)\n")

# Preprocessing Recipes
# =========================================================================

recipe_normalized <- 
  recipe(default ~ ., data = credit_train) %>%
  step_dummy(all_nominal_predictors()) %>%
  step_zv(all_predictors()) %>%
  step_normalize(all_numeric_predictors())

cat("✓ Normalized recipe created\n")
cat("Steps:\n")
cat("  1. Create dummy variables for categorical predictors\n")
cat("  2. Remove zero-variance predictors\n")
cat("  3. Normalize numeric predictors\n")
cat("Used for: Logistic, Lasso, Ridge, KNN\n")

recipe_basic <- 
  recipe(default ~ ., data = credit_train) %>%
  step_dummy(all_nominal_predictors()) %>%
  step_zv(all_predictors())

cat("✓ Basic recipe created\n")
cat("Steps:\n")
cat("  1. Create dummy variables for categorical predictors\n")
cat("  2. Remove zero-variance predictors\n")
cat("Used for: Decision Tree, Random Forest\n")

# Workflows
# =========================================================================

# Logistic Regression Workflow
logistic_wf <- 
  workflow() %>%
  add_model(logistic_spec) %>%
  add_recipe(recipe_normalized)

# Lasso Regression Workflow
lasso_wf <- 
  workflow() %>%
  add_model(lasso_spec) %>%
  add_recipe(recipe_normalized)

# Ridge Regression Workflow
ridge_wf <- 
  workflow() %>%
  add_model(ridge_spec) %>%
  add_recipe(recipe_normalized)

# Decision Tree Workflow
tree_wf <- 
  workflow() %>%
  add_model(tree_spec) %>%
  add_recipe(recipe_basic)

# Random Forest Workflow
rf_wf <- 
  workflow() %>%
  add_model(rf_spec) %>%
  add_recipe(recipe_basic)

# KNN Workflow
knn_wf <- 
  workflow() %>%
  add_model(knn_spec) %>%
  add_recipe(recipe_normalized)

cat("✓ All 6 workflows created successfully\n")
cat("Each workflow combines:\n")
cat("  - Model specification\n")
cat("  - Preprocessing recipe\n")
cat("  - Ready for training and tuning\n")

# Tuning Grids
# =========================================================================

# Define parameter grids
lasso_grid <- grid_regular(penalty(), levels = 20)
ridge_grid <- grid_regular(penalty(), levels = 20)

tree_grid <- grid_regular(
  cost_complexity(),
  tree_depth(),
  min_n(),
  levels = 5
)

rf_grid <- grid_regular(
  mtry(range = c(2, 8)),
  min_n(),
  levels = 5
)

knn_grid <- grid_regular(
  neighbors(range = c(3, 20)),
  weight_func(),
  dist_power(),
  levels = 5
)

cat("Tuning grids created:\n")
cat("✓ Lasso: 20 penalty values\n")
cat("✓ Ridge: 20 penalty values\n")
cat("✓ Decision Tree: 5³ = 125 combinations\n")
cat("✓ Random Forest: 5² = 25 combinations\n")
cat("✓ KNN: 5³ = 125 combinations\n")

# Model Training and Tuning
# =========================================================================

# Metrics to track - fix namespace conflicts
class_metrics <- metric_set(yardstick::roc_auc, yardstick::accuracy, yardstick::precision, yardstick::recall, yardstick::f_meas)

cat("Starting model training with 5-fold cross-validation...\n\n")

# Tune Logistic Regression (no tuning needed for basic logistic)
cat("1/6 Training Logistic Regression...\n")
logistic_tune <- 
  logistic_wf %>%
  fit_resamples(
    resamples = credit_folds,
    metrics = class_metrics,
    control = control_resamples(save_pred = TRUE)
  )

# Tune Lasso Regression
cat("2/6 Training Lasso Regression...\n")
lasso_tune <- 
  lasso_wf %>%
  tune_grid(
    resamples = credit_folds,
    grid = lasso_grid,
    metrics = class_metrics,
    control = control_grid(save_pred = TRUE)
  )

# Tune Ridge Regression
cat("3/6 Training Ridge Regression...\n")
ridge_tune <- 
  ridge_wf %>%
  tune_grid(
    resamples = credit_folds,
    grid = ridge_grid,
    metrics = class_metrics,
    control = control_grid(save_pred = TRUE)
  )

# Tune Decision Tree
cat("4/6 Training Decision Tree...\n")
tree_tune <- 
  tree_wf %>%
  tune_grid(
    resamples = credit_folds,
    grid = tree_grid,
    metrics = class_metrics,
    control = control_grid(save_pred = TRUE)
  )

# Tune Random Forest
cat("5/6 Training Random Forest...\n")
rf_tune <- 
  rf_wf %>%
  tune_grid(
    resamples = credit_folds,
    grid = rf_grid,
    metrics = class_metrics,
    control = control_grid(save_pred = TRUE)
  )

# Tune KNN
cat("6/6 Training KNN...\n")
knn_tune <- 
  knn_wf %>%
  tune_grid(
    resamples = credit_folds,
    grid = knn_grid,
    metrics = class_metrics,
    control = control_grid(save_pred = TRUE)
  )

cat("\n✓ All models trained successfully!\n")

# Model Finalization
# =========================================================================

# Select best hyperparameters for each model
lasso_best_params <- lasso_tune %>% select_best(metric = "roc_auc")
ridge_best_params <- ridge_tune %>% select_best(metric = "roc_auc")
tree_best_params <- tree_tune %>% select_best(metric = "roc_auc")
rf_best_params <- rf_tune %>% select_best(metric = "roc_auc")
knn_best_params <- knn_tune %>% select_best(metric = "roc_auc")

cat("Best hyperparameters selected:\n")
cat("✓ Lasso penalty:", round(lasso_best_params$penalty, 6), "\n")
cat("✓ Ridge penalty:", round(ridge_best_params$penalty, 6), "\n")
cat("✓ Tree complexity:", round(tree_best_params$cost_complexity, 6), "\n")
cat("✓ RF mtry:", rf_best_params$mtry, "\n")
cat("✓ KNN neighbors:", knn_best_params$neighbors, "\n")

# Finalize workflows
logistic_final <- logistic_wf
lasso_final <- lasso_wf %>% finalize_workflow(lasso_best_params)
ridge_final <- ridge_wf %>% finalize_workflow(ridge_best_params)
tree_final <- tree_wf %>% finalize_workflow(tree_best_params)
rf_final <- rf_wf %>% finalize_workflow(rf_best_params)
knn_final <- knn_wf %>% finalize_workflow(knn_best_params)

# Last fit on test set
cat("\nFitting final models on test set...\n")
logistic_last_fit <- logistic_final %>% last_fit(credit_split, metrics = class_metrics)
lasso_last_fit <- lasso_final %>% last_fit(credit_split, metrics = class_metrics)
ridge_last_fit <- ridge_final %>% last_fit(credit_split, metrics = class_metrics)
tree_last_fit <- tree_final %>% last_fit(credit_split, metrics = class_metrics)
rf_last_fit <- rf_final %>% last_fit(credit_split, metrics = class_metrics)
knn_last_fit <- knn_final %>% last_fit(credit_split, metrics = class_metrics)

cat("✓ All final models fitted on test set!\n")

# Cross-Validation Results
# =========================================================================

# Collect best results for each model
logistic_best <- logistic_tune %>% collect_metrics() %>% filter(.metric == "roc_auc") %>% slice_max(mean, n = 1)
lasso_best <- lasso_tune %>% show_best(metric = "roc_auc", n = 1)
ridge_best <- ridge_tune %>% show_best(metric = "roc_auc", n = 1)
tree_best <- tree_tune %>% show_best(metric = "roc_auc", n = 1)
rf_best <- rf_tune %>% show_best(metric = "roc_auc", n = 1)
knn_best <- knn_tune %>% show_best(metric = "roc_auc", n = 1)

# Combine results
cv_results <- bind_rows(
  logistic_best %>% mutate(model = "Logistic Regression"),
  lasso_best %>% mutate(model = "Lasso Regression"),
  ridge_best %>% mutate(model = "Ridge Regression"),
  tree_best %>% mutate(model = "Decision Tree"),
  rf_best %>% mutate(model = "Random Forest"),
  knn_best %>% mutate(model = "KNN")
) %>%
  select(model, .metric, mean, std_err) %>%
  arrange(desc(mean))

# Display results
print(knitr::kable(cv_results, digits = 4, caption = "Cross-Validation Performance (ROC-AUC)"))

# ROC Curves Analysis
# =========================================================================

# Collect predictions for ROC curves
logistic_pred <- logistic_last_fit %>% collect_predictions()
lasso_pred <- lasso_last_fit %>% collect_predictions()
ridge_pred <- ridge_last_fit %>% collect_predictions()
tree_pred <- tree_last_fit %>% collect_predictions()
rf_pred <- rf_last_fit %>% collect_predictions()
knn_pred <- knn_last_fit %>% collect_predictions()

# Combine predictions
all_preds <- bind_rows(
  logistic_pred %>% mutate(model = "Logistic Regression"),
  lasso_pred %>% mutate(model = "Lasso Regression"),
  ridge_pred %>% mutate(model = "Ridge Regression"),
  tree_pred %>% mutate(model = "Decision Tree"),
  rf_pred %>% mutate(model = "Random Forest"),
  knn_pred %>% mutate(model = "KNN")
)

# Plot ROC curves
roc_plot <- all_preds %>%
  group_by(model) %>%
  roc_curve(default, .pred_0) %>%
  ggplot(aes(x = 1 - specificity, y = sensitivity, color = model)) +
  geom_line(size = 1.2) +
  geom_abline(lty = 2, alpha = 0.5) +
  labs(title = "ROC Curves Comparison",
       subtitle = "Diagonal line represents random classifier",
       x = "1 - Specificity (False Positive Rate)",
       y = "Sensitivity (True Positive Rate)",
       color = "Model") +
  theme_minimal() +
  theme(legend.position = "bottom",
        plot.title = element_text(hjust = 0.5),
        plot.subtitle = element_text(hjust = 0.5))

print(roc_plot)

# Test Set Performance
# =========================================================================

# Collect test set metrics
test_results <- bind_rows(
  logistic_last_fit %>% collect_metrics() %>% mutate(model = "Logistic Regression"),
  lasso_last_fit %>% collect_metrics() %>% mutate(model = "Lasso Regression"),
  ridge_last_fit %>% collect_metrics() %>% mutate(model = "Ridge Regression"),
  tree_last_fit %>% collect_metrics() %>% mutate(model = "Decision Tree"),
  rf_last_fit %>% collect_metrics() %>% mutate(model = "Random Forest"),
  knn_last_fit %>% collect_metrics() %>% mutate(model = "KNN")
) %>%
  select(model, .metric, .estimate) %>%
  pivot_wider(names_from = .metric, values_from = .estimate) %>%
  arrange(desc(roc_auc))

# Display test results
print(knitr::kable(test_results, digits = 4, caption = "Test Set Performance - All Metrics"))

# Summary
# =========================================================================
cat("\n===============================================\n")
cat("ANALYSIS COMPLETE\n")
cat("===============================================\n")
cat("Best performing model on test set:", test_results$model[1], "\n")
cat("ROC-AUC:", round(test_results$roc_auc[1], 4), "\n")
cat("Accuracy:", round(test_results$accuracy[1], 4), "\n")
cat("===============================================\n") 